(defpackage #:delsegno.gc-bench
  (:use #:common-lisp))

(in-package #:delsegno.gc-bench)

(defstruct node left right dummy1 dummy2)

;;; build tree top down, assigning to older objects
(defun populate (depth node)
  (when (> depth 0)
    (setf (node-left node) (make-node))
    (setf (node-right node) (make-node))
    (populate (1- depth) (node-left node))
    (populate (1- depth) (node-right node))))

;;; build tree bottom-up
(defun make-tree (depth)
  (if (<= depth 0)
      (make-node)
      (make-node :left (make-tree (- depth 1))
                 :right (make-tree (- depth 1)))))

;;; nodes used by a tree of a given size
(defmacro tree-size (i)
  `(- (expt 2 (1+ ,i)) 1))

;;; number of iterations to use for a given tree depth
(defmacro iteration-count (i)
  `(floor (* 2 (tree-size stretch-tree-depth))
          (tree-size ,i)))

(defun gcbench (stretch-tree-depth)
  (declare (fixnum stretch-tree-depth))
  ;;   Parameters are determined by stretch-tree-depth.
  ;;   In Boehm's version the parameters were fixed as follows:
  ;;     public static final int stretch-tree-depth    = 18;  // about 16Mb
  ;;     public static final int kLongLivedTreeDepth  = 16;  // about 4Mb
  ;;     public static final int kArraySize  = 500000;       // about 4Mb
  ;;     public static final int min-tree-depth = 4;
  ;;     public static final int kMaxTreeDepth = 16;
  ;;   In Larceny the storage numbers above would be 12 Mby, 3 Mby, 6 Mby.
  (let* ((long-lived-tree-depth (- stretch-tree-depth 2))
         (array-size          (* 4 (tree-size long-lived-tree-depth)))
         (min-tree-depth       4)
         (max-tree-depth       long-lived-tree-depth))
    ;; (format t "Stretching memory with a binary tree of depth ~d~%"
    ;;         stretch-tree-depth)
    ;; stretch the memory space quickly
    (make-tree stretch-tree-depth)
    ;; Create a long lived object
    ;; (format t "Creating a long-lived binary tree of depth ~d~%"
    ;;           long-lived-tree-depth)
    (let ((long-lived-tree (make-node)))
      (populate long-lived-tree-depth long-lived-tree)
      ;; create long-lived array, filling half of it
      ;; (format t "Creating a long-lived array of ~d inexact reals~%"
      ;;         array-size)
      (let ((array (make-array array-size :element-type 'single-float)))
        (loop for i below (floor array-size 2)
              do (setf (aref array i) (/ 1.0 (1+ i))))
        (loop for d from min-tree-depth upto max-tree-depth by 2 do
          (let ((iteration-count (iteration-count d)))
            ;; (format t "~&Creating ~d trees of depth ~d~%"
            ;;         iteration-count d)
            ;; (format t "GCBench: Top down construction~%")
            (dotimes (i iteration-count) (populate d (make-node)))
            ;; (format t "GCBench: Bottom up construction~%")
            (dotimes (i iteration-count) (make-tree d))))
        ;; these are fake references to LongLivedTree and array to
        ;; keep them from being optimized away
        (assert (not (null long-lived-tree)))
        (assert (let ((n (min 1000 (1- (floor (length array) 2)))))
                  (= (round (aref array n)) (round (/ 1.0 (1+ n))))))))))

;; default K value was originally 18.
(delsegno:define-benchmark gc-bench (&optional (k 20))
  ;; (format t "The garbage collector should touch about ~D megabytes ~"
  ;;           "of heap storage.~%" (expt 2 (- k 13)))
  ;; (format t "The use of more or less memory will skew the results.~%")
  (gcbench k))

(delsegno:define-benchmark threaded-gc-bench (&optional (n-threads 2) (k 19))
  #+sb-thread
  (loop repeat n-threads
        collect (sb-thread:make-thread (lambda () (gcbench k))) into threads
        finally (mapc #'sb-thread:join-thread threads))
  #-sb-thread
  (error "Not implemented."))